@@ -4,7 +4,7 @@ from dataclasses import dataclass
 from enum import Enum
 from functools import cached_property
 from math import ceil, floor
-from typing import Type
+from typing import Sequence, Type
 
 # Local imports
 from . import validators
@@ -52,7 +52,12 @@ class Side(str, Enum):
     @classmethod
     def from_string(cls, side_string: str) -> Side:
         # use a lookup to ensure KeyError with unknown value
-        return {'short': cls.SHORT, 'long': cls.LONG}[side_string.lower()]
+        lookup = {
+            'short': cls.SHORT,
+            'hedge': cls.SHORT,
+            'long': cls.LONG,
+        }
+        return lookup[side_string.lower()]
 
 
 class NominalOracleUnitsXSatsPerBch(int):
@@ -87,6 +92,7 @@ class LongLeverage(float):
 class RedemptionType(str, Enum):
     LIQUIDATION = 'Liquidation'
     MATURATION = 'Maturation'
+    MUTUAL = 'Mutual'
 
     def __str__(self):
         return self.value
@@ -96,20 +102,29 @@ class RedemptionType(str, Enum):
 
     @classmethod
     def from_string(cls, redemption_type_string: str) -> RedemptionType:
-        # use a lookup to ensure KeyError with unknown value
-        return {'liquidation': cls.LIQUIDATION, 'maturation': cls.MATURATION}[redemption_type_string.lower()]
+        # use a strict lookup (with lowercase) to ensure KeyError with unknown value
+        lookup = {
+            'liquidation': cls.LIQUIDATION,
+            'maturation': cls.MATURATION,
+            'mutual': cls.MUTUAL,
+        }
+        redemption_type = lookup[redemption_type_string.lower()]
+        return redemption_type
 
 
 @dataclass(frozen=True)
 class ContractProposal:
     """Details of a proposed contract between a maker and taker. Does not include any funding oriented details such as fees."""
+    # Address (unvalidated, use empty string when unknown)
+    address: str
+
     # Time
     start_timestamp: ScriptTimestamp
     maturity_timestamp: ScriptTimestamp
 
     # Position settings
     nominal_oracleUnits_x_satsPerBch: NominalOracleUnitsXSatsPerBch
-    cost_sats_for_nominal_value_at_short_liquidation: Sats
+    cost_sats_for_nominal_value_at_high_liquidation: Sats
     total_input_sats: UtxoSats
 
     # Start price is not an actual contract parameter, but one of start price, leverage,
@@ -117,8 +132,8 @@ class ContractProposal:
     start_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch
 
     # Liquidation prices
-    short_liquidation_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch
-    long_liquidation_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch
+    high_liquidation_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch
+    low_liquidation_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch
 
     # Price oracle
     oracle_public_key: PublicKey
@@ -141,7 +156,7 @@ class ContractProposal:
         # The definition of short's input is that it needs to be enough to fully cover the worst case outcome of selling low
         # and buying high. That amount is the difference between the cost of the nominal assets at short/high liquidation vs
         # the cost at the start price. Note that in our inverted price scheme of BCH/Asset, the lower cost is at the higher price.
-        short_input_sats = UtxoSats(self.cost_sats_for_nominal_value_at_start - self.cost_sats_for_nominal_value_at_short_liquidation)
+        short_input_sats = UtxoSats(self.cost_sats_for_nominal_value_at_start - self.cost_sats_for_nominal_value_at_high_liquidation)
 
         # Long input sats is everything needed to cover the rest of the total
         long_input_sats = UtxoSats(self.total_input_sats - short_input_sats)
@@ -187,37 +202,49 @@ class ContractProposal:
 
     @property
     def cost_sats_for_nominal_value_at_start(self) -> UtxoSats:
-        return UtxoSats(round_half_up((self.effective_nominal_value_oracleUnits / float(self.start_price_oracleUnits_per_bch)) * SATS_PER_BCH))
+        return UtxoSats(round_half_up(float(self.nominal_oracleUnits_x_satsPerBch) / float(self.start_price_oracleUnits_per_bch)))
 
     @property
     def effective_short_leverage(self) -> ShortLeverage:
         # There is a special case where even if the liquidation price must be recorded at some max value,
         # The fundamental cost for the position at liquidation is hard-lined at zero. That is the definition
         # of a hard hedge position, emulating the original anyhedge contract behavior.
-        if self.cost_sats_for_nominal_value_at_short_liquidation == 0:
+        if self.cost_sats_for_nominal_value_at_high_liquidation == 0:
             return ShortLeverage(1)
 
         # Derivation of the calculation:
-        #   short liq price = start price (1 + 1 / (short leverage - 1))
+        #   short liq price = high liq price = start price (1 + 1 / (short leverage - 1))
         #   1 + 1 / (short leverage - 1) = short liq price / start price
         #   1 / (short leverage - 1) = (short liq price / start price) - 1
         #   short leverage - 1 = 1 / ((short liq price / start price) - 1)
         #   short leverage = 1 + 1 / ((short liq price / start price) - 1)
-        return ShortLeverage(1 + 1 / (self.short_liquidation_price_oracleUnits_per_bch / self.start_price_oracleUnits_per_bch - 1))
+        return ShortLeverage(1 + 1 / (self.high_liquidation_price_oracleUnits_per_bch / self.start_price_oracleUnits_per_bch - 1))
 
     @property
     def effective_long_leverage(self) -> LongLeverage:
         # Derivation of the calculation:
-        #   long liq price = start price (1 - 1 / long leverage)
+        #   long liq price = low liq price = start price (1 - 1 / long leverage)
         #   1 - 1 / long leverage = long liq price / start price
         #   1 / long leverage = 1 - (long liq price / start price)
         #   long leverage = 1 / (1 - (long liq price / start price))
-        return LongLeverage(1 / (1 - self.long_liquidation_price_oracleUnits_per_bch / self.start_price_oracleUnits_per_bch))
+        return LongLeverage(1 / (1 - self.low_liquidation_price_oracleUnits_per_bch / self.start_price_oracleUnits_per_bch))
 
     @property
     def taker_side(self) -> Side:
         return self.maker_side.other_side
 
+    @property
+    def short_role(self) -> Role:
+        if self.maker_side == Side.SHORT:
+            return Role.MAKER
+        return Role.TAKER
+
+    @property
+    def long_role(self) -> Role:
+        if self.maker_side == Side.LONG:
+            return Role.MAKER
+        return Role.TAKER
+
     ###############
     # Property access to input sats
     ###############
@@ -272,54 +299,82 @@ class ContractProposal:
                         maker_side: Side,
                         # For upgrade to leveraged shorts, we establish a default of 1 representing the pure hedge position
                         short_leverage: ShortLeverage = ShortLeverage(1),
+                        address: str = '',
                         ) -> ContractProposal:
         nominal_oracleUnits_x_satsPerBch = NominalOracleUnitsXSatsPerBch(round_half_up(nominal_oracleUnits * SATS_PER_BCH))
 
         try:
-            # Short liquidation price is a hard contract parameter, and is also used as a base number to derive other contract details.
-            # We calculate short liquidation price from the more abstract leverage input and then discard leverage.
-            # Short liquidation price is rounded for accuracy - in the contract it's used as for testing, but not calculations
-            short_liquidation_price_oracleUnits_per_bch = ScriptPriceInOracleUnitsPerBch(round_half_up(float(start_price_oracleUnits_per_bch) * (1 + 1 / (short_leverage - 1))))
+            # High liquidation price is a hard contract parameter, and is also used as a base number to derive other contract details.
+            # We calculate high liquidation price from the more abstract leverage input and then discard leverage.
+            # High liquidation price is rounded for accuracy - in the contract it's used as for testing, but not calculations
+            high_liquidation_price_oracleUnits_per_bch = ScriptPriceInOracleUnitsPerBch(round_half_up(float(start_price_oracleUnits_per_bch) * (1 + 1 / (short_leverage - 1))))
 
             # Each side of the contract has a worst-case price (the liquidation price) at which it needs to be able to settle the contract.
             # In nominal terms, there is a full cost for that position, but in input/output terms, the counterparties only need to cover the difference in cost.
-            # In order to achieve the simplest possible in-contract calculation, we use the full cost for short liquidation price as a contract parameter.
+            # In order to achieve the simplest possible in-contract calculation, we use the full cost for high liquidation price as a contract parameter.
             # cost is floored for safety - it's a hard contract value subtracted from another during contract execution, and that result must never be a negative number
-            cost_sats_for_nominal_value_at_short_liquidation = Sats(floor(nominal_oracleUnits_x_satsPerBch / short_liquidation_price_oracleUnits_per_bch))
+            cost_sats_for_nominal_value_at_high_liquidation = Sats(floor(nominal_oracleUnits_x_satsPerBch / high_liquidation_price_oracleUnits_per_bch))
         except ZeroDivisionError:
             # For short leverage exactly 1 (or any floating point result that still causes the error), the implication
-            # is that short liquidation is at infinity on a continuous number line. However, for the contract purposes,
+            # is that high liquidation is at infinity on a continuous number line. However, for the contract purposes,
             # we need a concrete test price for liquidation and the cost value. We establish those directly:
-            short_liquidation_price_oracleUnits_per_bch = ScriptPriceInOracleUnitsPerBch.largest_allowed()
-            cost_sats_for_nominal_value_at_short_liquidation = Sats(0)
+            high_liquidation_price_oracleUnits_per_bch = ScriptPriceInOracleUnitsPerBch.largest_allowed()
+            cost_sats_for_nominal_value_at_high_liquidation = Sats(0)
 
-        # Long liquidation has the same considerations as short liquidation above, except long leverage of 1 is strictly disallowed,
+        # Low liquidation has the same considerations as high liquidation above, except long leverage of 1 is strictly disallowed,
         # avoiding the problem of a zero price, infinite cost. Note that a short leverage of 1 is useful in the sense that it gives the short
         # party the equivalent of simply holding another asset, while long leverage of 1 is nonsensical in that it gives the long
         # the equivalent of simply holding the underlying asset (BCH) which by definition they already have.
-        long_liquidation_price_oracleUnits_per_bch = ScriptPriceInOracleUnitsPerBch(round_half_up(float(start_price_oracleUnits_per_bch) * (1 - 1 / long_leverage)))
+        low_liquidation_price_oracleUnits_per_bch = ScriptPriceInOracleUnitsPerBch(round_half_up(float(start_price_oracleUnits_per_bch) * (1 - 1 / long_leverage)))
 
-        # Cost at long liquidation has the same considerations as short liquidation above, except we are on the opposite side so we use ceil to be conservative instead of floor
-        cost_sats_for_nominal_value_at_long_liquidation = UtxoSats(ceil(nominal_oracleUnits_x_satsPerBch / long_liquidation_price_oracleUnits_per_bch))
+        # Cost at low liquidation has the same considerations as high liquidation above, except we are on the opposite side so we use ceil to be conservative instead of floor
+        cost_sats_for_nominal_value_at_low_liquidation = UtxoSats(ceil(nominal_oracleUnits_x_satsPerBch / low_liquidation_price_oracleUnits_per_bch))
 
         # Total input sats is the total of the two party's required amounts. In the contract base terms that we are using, it is the difference between
         # the cost at the two liquidation points
-        total_input_sats = UtxoSats(cost_sats_for_nominal_value_at_long_liquidation - cost_sats_for_nominal_value_at_short_liquidation)
+        total_input_sats = UtxoSats(cost_sats_for_nominal_value_at_low_liquidation - cost_sats_for_nominal_value_at_high_liquidation)
 
         return ContractProposal(
+            address=address,
             start_timestamp=start_timestamp,
             maturity_timestamp=maturity_timestamp,
             nominal_oracleUnits_x_satsPerBch=nominal_oracleUnits_x_satsPerBch,
-            cost_sats_for_nominal_value_at_short_liquidation=cost_sats_for_nominal_value_at_short_liquidation,
+            cost_sats_for_nominal_value_at_high_liquidation=cost_sats_for_nominal_value_at_high_liquidation,
             total_input_sats=total_input_sats,
             start_price_oracleUnits_per_bch=start_price_oracleUnits_per_bch,
-            short_liquidation_price_oracleUnits_per_bch=short_liquidation_price_oracleUnits_per_bch,
-            long_liquidation_price_oracleUnits_per_bch=long_liquidation_price_oracleUnits_per_bch,
+            high_liquidation_price_oracleUnits_per_bch=high_liquidation_price_oracleUnits_per_bch,
+            low_liquidation_price_oracleUnits_per_bch=low_liquidation_price_oracleUnits_per_bch,
             oracle_public_key=nominal_oracleUnits.public_key,
             maker_side=maker_side,
         )
 
-    def fund(self, fee_agreements: list[FeeAgreement]) -> ContractFunding:
+    def neutralize(self,
+                   current_price_oracleUnitsPerBch: ScriptPriceInOracleUnitsPerBch,
+                   current_timestamp: ScriptTimestamp,
+                   ) -> ContractProposal:
+        neutralizing_proposal = ContractProposal(
+            # reset / unimplemented
+            address='',
+
+            # update to settlement timing / price
+            start_timestamp=current_timestamp,
+            start_price_oracleUnits_per_bch=current_price_oracleUnitsPerBch,
+
+            # swap sides
+            maker_side=self.taker_side,
+
+            # the rest remains the same
+            maturity_timestamp=self.maturity_timestamp,
+            nominal_oracleUnits_x_satsPerBch=self.nominal_oracleUnits_x_satsPerBch,
+            cost_sats_for_nominal_value_at_high_liquidation=self.cost_sats_for_nominal_value_at_high_liquidation,
+            total_input_sats=self.total_input_sats,
+            high_liquidation_price_oracleUnits_per_bch=self.high_liquidation_price_oracleUnits_per_bch,
+            low_liquidation_price_oracleUnits_per_bch=self.low_liquidation_price_oracleUnits_per_bch,
+            oracle_public_key=self.oracle_public_key,
+        )
+        return neutralizing_proposal
+
+    def fund(self, fee_agreements: Sequence[FeeAgreement]) -> ContractFunding:
         return ContractFunding(
             base_proposal=self,
             fee_agreements=tuple(fee_agreements),
@@ -392,33 +447,39 @@ class ContractFunding:
                price_timestamp: ScriptTimestamp,
                price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch,
                force_maturity: bool,
+               is_mutual_redemption: bool = False,
+               fee_agreements: Sequence[FeeAgreement] = tuple(),
                ) -> ContractRedemption:
-        """Redeem the contract according to market conditions or raise an unredeemable error for invalid conditions."""
+        """
+        Redeem the contract according to market conditions or raise an unredeemable error for invalid conditions.
+        Note that is_mutual_redemption takes precedence over force_maturity.
+        """
         reached_maturity_time = price_timestamp >= self.base_proposal.maturity_timestamp
         reached_liquidation_price = (
-                price_oracleUnits_per_bch <= self.base_proposal.long_liquidation_price_oracleUnits_per_bch
+                price_oracleUnits_per_bch <= self.base_proposal.low_liquidation_price_oracleUnits_per_bch
                 or
-                price_oracleUnits_per_bch >= self.base_proposal.short_liquidation_price_oracleUnits_per_bch
+                price_oracleUnits_per_bch >= self.base_proposal.high_liquidation_price_oracleUnits_per_bch
         )
-        if reached_maturity_time or force_maturity:
+
+        if is_mutual_redemption:
+            # Mutual redemption
+            redemption_type = RedemptionType.MUTUAL
+        elif reached_maturity_time or force_maturity:
             # Maturation, even in the case of a liquidation price
-            return ContractRedemption(
-                base_funding=self,
-                end_price_timestamp=price_timestamp,
-                naive_end_price_oracleUnits_per_bch=price_oracleUnits_per_bch,
-                redemption_type=RedemptionType.MATURATION,
-            )
+            redemption_type = RedemptionType.MATURATION
         elif reached_liquidation_price:
             # Liquidation
-            return ContractRedemption(
-                base_funding=self,
-                end_price_timestamp=price_timestamp,
-                naive_end_price_oracleUnits_per_bch=price_oracleUnits_per_bch,
-                redemption_type=RedemptionType.LIQUIDATION,
-            )
+            redemption_type = RedemptionType.LIQUIDATION
+        else:
+            raise UnredeemableError
 
-        # If conditions did not allow for redemption, throw an error
-        raise UnredeemableError
+        return ContractRedemption(
+            base_funding=self,
+            end_price_timestamp=price_timestamp,
+            naive_end_price_oracleUnits_per_bch=price_oracleUnits_per_bch,
+            redemption_type=redemption_type,
+            fee_agreements=tuple(fee_agreements),
+        )
 
 
 @dataclass(frozen=True)
@@ -428,6 +489,7 @@ class ContractRedemption:
     end_price_timestamp: ScriptTimestamp
     naive_end_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch
     redemption_type: RedemptionType
+    fee_agreements: tuple[FeeAgreement, ...]
 
     ###############
     # Parameterized Payout Values
@@ -442,7 +504,7 @@ class ContractRedemption:
 
         # With leveraged shorts, the short is no longer necessarily paying out the full value of the nominal position.
         # If short leverage is not 1, the short only pays up to the planned liquidation point
-        unsafe_short_payout_sats = _unsafe_hedge_payout_sats - self.base_funding.base_proposal.cost_sats_for_nominal_value_at_short_liquidation
+        unsafe_short_payout_sats = _unsafe_hedge_payout_sats - self.base_funding.base_proposal.cost_sats_for_nominal_value_at_high_liquidation
         short_payout_sats = UtxoSats(max(DUST, unsafe_short_payout_sats))
 
         # Long Payout Sats
@@ -488,7 +550,11 @@ class ContractRedemption:
     ###############
     @property
     def clamped_end_price_oracleUnits_per_bch(self) -> ScriptPriceInOracleUnitsPerBch:
-        return max(self.naive_end_price_oracleUnits_per_bch, self.base_funding.base_proposal.long_liquidation_price_oracleUnits_per_bch)
+        return max(self.naive_end_price_oracleUnits_per_bch, self.base_funding.base_proposal.low_liquidation_price_oracleUnits_per_bch)
+
+    @property
+    def cost_sats_for_nominal_value_at_redemption(self) -> Sats:
+        return Sats(round_half_up(SATS_PER_BCH * (self.base_funding.base_proposal.effective_nominal_value_oracleUnits / float(self.clamped_end_price_oracleUnits_per_bch))))
 
     ###############
     # Property access to payout sats
@@ -621,3 +687,55 @@ class ContractRedemption:
         if self.base_funding.base_proposal.taker_side == Side.SHORT:
             return self.short_gain_percent_of_own_input
         return self.long_gain_percent_of_own_input
+
+    ###############
+    # Redemption Fees (Especially for early redemption)
+    ###############
+    @property
+    def fee_sats_to_maker(self) -> Sats:
+        return aggregate_fee_sats_to_role(self.fee_agreements, Role.MAKER)
+
+    @property
+    def fee_sats_to_taker(self) -> Sats:
+        return aggregate_fee_sats_to_role(self.fee_agreements, Role.TAKER)
+
+    @property
+    def fee_sats_to_short(self) -> Sats:
+        if self.base_funding.base_proposal.maker_side == Side.SHORT:
+            return self.fee_sats_to_maker
+        return self.fee_sats_to_taker
+
+    @property
+    def fee_sats_to_long(self) -> Sats:
+        if self.base_funding.base_proposal.maker_side == Side.LONG:
+            return self.fee_sats_to_maker
+        return self.fee_sats_to_taker
+
+    ###############
+    # Unit value of fees
+    ###############
+    @property
+    def fee_oracleUnits_to_maker(self) -> OracleUnit:
+        fee_bch = self.fee_sats_to_maker.bch
+        # NOTE: using actual end price, not clamped, to determine unit value including any potential slippage
+        fee_oracleUnits = self.base_funding.base_proposal.oracle_unit_cls(fee_bch * float(self.naive_end_price_oracleUnits_per_bch))
+        return fee_oracleUnits
+
+    @property
+    def fee_oracleUnits_to_taker(self) -> OracleUnit:
+        fee_bch = self.fee_sats_to_taker.bch
+        # NOTE: using actual end price, not clamped, to determine unit value including any potential slippage
+        fee_oracleUnits = self.base_funding.base_proposal.oracle_unit_cls(fee_bch * float(self.naive_end_price_oracleUnits_per_bch))
+        return fee_oracleUnits
+
+    @property
+    def fee_oracleUnits_to_short(self) -> OracleUnit:
+        if self.base_funding.base_proposal.maker_side == Side.SHORT:
+            return self.fee_oracleUnits_to_maker
+        return self.fee_oracleUnits_to_taker
+
+    @property
+    def fee_oracleUnits_to_long(self) -> OracleUnit:
+        if self.base_funding.base_proposal.maker_side == Side.LONG:
+            return self.fee_oracleUnits_to_maker
+        return self.fee_oracleUnits_to_taker
