@@ -1,5 +1,6 @@
 # Built-in imports
 import unittest
+from typing import Sequence
 
 # Library imports
 from arrow import Arrow
@@ -14,6 +15,7 @@ from ..bch_primitives import (
 from ..contract import (
     ContractFunding,
     ContractProposal,
+    ContractRedemption,
     LongLeverage,
     NominalOracleUnitsXSatsPerBch,
     RedemptionType,
@@ -59,47 +61,47 @@ NOMINAL_UNITS_X_SATS_PER_BCH = NominalOracleUnitsXSatsPerBch(1_000_000_000_000)
 # Calculated as NOMINAL_UNITS_X_SATS_PER_BCH / SATS_PER_BCH
 EFFECTIVE_NOMINAL_VALUE_ORALCEUNITS = UsdEM2Beta(100_00.0)
 
-# Long liquidation Price is the price where all of long's input is consumed to cover short's agreed payout
+# Low liquidation Price is the price where all of long's input is consumed to cover short's agreed payout
 # Calculate in terms of leverage intent as round(START_PRICE_USDEM2_PER_BCH * (1 - 1 / LONG_LEVERAGE))
 # 15556 cents/BCH ($155.56 / BCH)
-LONG_LIQUIDATION_PRICE = ScriptPriceInOracleUnitsPerBch(155_56)
+LOW_LIQUIDATION_PRICE = ScriptPriceInOracleUnitsPerBch(155_56)
 
-# Short liquidation Price is the price where all of short's input is consumed to cover long's agreed payout
+# High liquidation Price is the price where all of short's input is consumed to cover long's agreed payout
 # When short is exactly a 1x short, there is no meaningful liquidation price (infinity) but the stored liquidation price is still
 # used to trigger early settlement. By definition, we set it to the max possible price in the 4 byte oracle price message scheme.
-SHORT_LIQUIDATION_PRICE_AS_HEDGE = ScriptPriceInOracleUnitsPerBch.largest_allowed()
+HIGH_LIQUIDATION_PRICE_AS_HEDGE = ScriptPriceInOracleUnitsPerBch.largest_allowed()
 # In the case of a leveraged short, it is calculated as round(START_PRICE_USDEM2_PER_BCH * (1 + 1 / (SHORT_LEVERAGE_NOT_1 - 1)))
 # 36667 cents/BCH ($366.67 / BCH)
-SHORT_LIQUIDATION_PRICE_AS_LEVERAGED_SHORT = ScriptPriceInOracleUnitsPerBch(366_67)
+HIGH_LIQUIDATION_PRICE_AS_LEVERAGED_SHORT = ScriptPriceInOracleUnitsPerBch(366_67)
 
-# Cost of nominal value at short liquidation is a new value that was always zero for pure hedges
+# Cost of nominal value at high liquidation is a new value that was always zero for pure hedges
 # and therefore not documented explicitly anywhere. With leveraged shorts, short does not always
 # pay the full cost of the nominal value and so the contract needs to track the cost.
 # Fixed for pure hedge at 0
 ZERO_COST_OF_NOMINAL_VALUE_AT_HEDGE_LIQUIDATION = Sats(0)
-# Calculated for leveraged short as floor(NOMINAL_UNITS_X_SATS_PER_BCH / SHORT_LIQUIDATION_PRICE_FOR_LEVERAGED_SHORT)
-COST_SATS_FOR_NOMINAL_VALUE_AT_LEVERAGED_SHORT_LIQUIDATION = Sats(27_272_479)
+# Calculated for leveraged short as floor(NOMINAL_UNITS_X_SATS_PER_BCH / HIGH_LIQUIDATION_PRICE_FOR_LEVERAGED_SHORT)
+COST_SATS_FOR_NOMINAL_VALUE_AT_LEVERAGED_HIGH_LIQUIDATION = Sats(27_272_479)
 
 # The bch amount to cover the short's payout obligations to long
 # For a pure hedge, it is the full cost of the nominal contract at starting price,
 # effectively allowing the price to increase to any value and still be able to pay out
 SHORT_INPUT_SATS_AS_HEDGE = COST_SATS_FOR_NOMINAL_VALUE_AT_START
-# For a leveraged short, it is the cost at starting price less the cost at short liquidation price
-# Calculated as SHORT_INPUT_SATS_AS_HEDGE - COST_SATS_FOR_NOMINAL_VALUE_AT_LEVERAGED_SHORT_LIQUIDATION
+# For a leveraged short, it is the cost at starting price less the cost at high liquidation price
+# Calculated as SHORT_INPUT_SATS_AS_HEDGE - COST_SATS_FOR_NOMINAL_VALUE_AT_LEVERAGED_HIGH_LIQUIDATION
 SHORT_INPUT_SATS_AS_LEVERAGED_SHORT = UtxoSats(22_727_521)
 
-# Similar to cost at short liquidation, there is also the cost at long liquidation.
+# Similar to cost at high liquidation, there is also the cost at low liquidation.
 # In the contract, this is not stored anywhere, but it is a useful intermediary for confirming tests.
-# It is the amount needed to cover the nominal unit value at long liquidation price
+# It is the amount needed to cover the nominal unit value at low liquidation price
 # Calculated as ceil((NOMINAL_UNITS_X_SATS_PER_BCH) / DERIVED_LIQUIDATION_PRICE_UNITS_PER_BCH)
-COST_SATS_FOR_NOMINAL_VALUE_AT_LONG_LIQUIDATION = UtxoSats(64_283_878)
+COST_SATS_FOR_NOMINAL_VALUE_AT_LOW_LIQUIDATION = UtxoSats(64_283_878)
 
 # The total input sats needed to cover all positions is the sum of the short and long inputs.
 # However, from base contract values, it is the difference between the cost at the two liquidation points
-# For a pure hedge, it has to cover the cost between long liquidation and zero
-TOTAL_INPUT_SATS_AS_HEDGE = COST_SATS_FOR_NOMINAL_VALUE_AT_LONG_LIQUIDATION - 0
-# For a leveraged short, it has to cover the cost between long liquidation and short liquidation (which also might be zero)
-# Calculated as COST_SATS_FOR_NOMINAL_VALUE_AT_LONG_LIQUIDATION - COST_SATS_FOR_NOMINAL_VALUE_AT_SHORT_LIQUIDATION
+# For a pure hedge, it has to cover the cost between low liquidation and zero
+TOTAL_INPUT_SATS_AS_HEDGE = COST_SATS_FOR_NOMINAL_VALUE_AT_LOW_LIQUIDATION - 0
+# For a leveraged short, it has to cover the cost between low liquidation and high liquidation (which also might be zero)
+# Calculated as COST_SATS_FOR_NOMINAL_VALUE_AT_LOW_LIQUIDATION - COST_SATS_FOR_NOMINAL_VALUE_AT_HIGH_LIQUIDATION
 TOTAL_INPUT_SATS_AS_LEVERAGED_SHORT = UtxoSats(37_011_399)
 
 # The bch amount to provide volatility protection as specified by the multiplier
@@ -145,14 +147,29 @@ class TestContractProposalDerivedValues(unittest.TestCase):
     def test_total_input_sats(self):
         self.assertEqual(standard_contract_proposal().total_input_sats, TOTAL_INPUT_SATS_AS_HEDGE)
 
-    def test_long_liquidation_price_oracleUnits_per_bch(self):
-        self.assertEqual(standard_contract_proposal().long_liquidation_price_oracleUnits_per_bch, LONG_LIQUIDATION_PRICE)
+    def test_low_liquidation_price_oracleUnits_per_bch(self):
+        self.assertEqual(standard_contract_proposal().low_liquidation_price_oracleUnits_per_bch, LOW_LIQUIDATION_PRICE)
 
-    def test_short_liquidation_price_oracleUnits_per_bch(self):
-        self.assertEqual(standard_contract_proposal().short_liquidation_price_oracleUnits_per_bch, SHORT_LIQUIDATION_PRICE_AS_HEDGE)
+    def test_high_liquidation_price_oracleUnits_per_bch(self):
+        self.assertEqual(standard_contract_proposal().high_liquidation_price_oracleUnits_per_bch, HIGH_LIQUIDATION_PRICE_AS_HEDGE)
 
-    def test_maker_side(self):
-        self.assertEqual(standard_contract_proposal().maker_side, Side.SHORT)
+    def test_maker_taker_side(self):
+        maker_short = standard_contract_proposal(maker_side=Side.SHORT)
+        self.assertEqual(maker_short.maker_side, Side.SHORT)
+        self.assertEqual(maker_short.taker_side, Side.LONG)
+
+        maker_long = standard_contract_proposal(maker_side=Side.LONG)
+        self.assertEqual(maker_long.maker_side, Side.LONG)
+        self.assertEqual(maker_long.taker_side, Side.SHORT)
+
+    def test_short_long_role(self):
+        maker_short = standard_contract_proposal(maker_side=Side.SHORT)
+        self.assertEqual(maker_short.short_role, Role.MAKER)
+        self.assertEqual(maker_short.long_role, Role.TAKER)
+
+        maker_long = standard_contract_proposal(maker_side=Side.LONG)
+        self.assertEqual(maker_long.long_role, Role.MAKER)
+        self.assertEqual(maker_long.short_role, Role.TAKER)
 
     ######################
     # Exact secondary values derived from contract parameters
@@ -218,6 +235,46 @@ class TestContractProposalDerivedValues(unittest.TestCase):
             taker_short_contract.input_sats(role=Role.MAKER, side=Side.SHORT)
 
 
+class TestContractProposalNeutralize(unittest.TestCase):
+    def test_neutralize_creates_an_opposite_virtual_hedge_at_current_price(self):
+        # Use a base case and hand-constructed neutralizing contract to confirm actual result
+
+        # Construct the base case
+        base_proposal = ContractProposal.new_from_intent(
+            start_timestamp=START_TIMESTAMP,
+            maturity_timestamp=MATURITY_TIMESTAMP,
+            nominal_oracleUnits=NOMINAL_VALUE_USDEM2,
+            long_leverage=LongLeverage(5.0),
+            start_price_oracleUnits_per_bch=START_PRICE_USDEM2_PER_BCH,
+            maker_side=Side.SHORT,
+            short_leverage=ShortLeverage(1.0),
+            address='',
+        )
+
+        # Manually construct the neutralizing contract
+        neutralizing_price = ScriptPriceInOracleUnitsPerBch(190_00)
+        neutralizing_timestamp = ScriptTimestamp(round((START_TIMESTAMP + MATURITY_TIMESTAMP) / 2))
+        expected_neutralizing_contract = ContractProposal(
+            address='',
+            start_timestamp=neutralizing_timestamp,
+            maturity_timestamp=base_proposal.maturity_timestamp,
+            nominal_oracleUnits_x_satsPerBch=base_proposal.nominal_oracleUnits_x_satsPerBch,
+            cost_sats_for_nominal_value_at_high_liquidation=base_proposal.cost_sats_for_nominal_value_at_high_liquidation,
+            total_input_sats=base_proposal.total_input_sats,
+            start_price_oracleUnits_per_bch=neutralizing_price,
+            high_liquidation_price_oracleUnits_per_bch=base_proposal.high_liquidation_price_oracleUnits_per_bch,
+            low_liquidation_price_oracleUnits_per_bch=base_proposal.low_liquidation_price_oracleUnits_per_bch,
+            oracle_public_key=base_proposal.oracle_public_key,
+            maker_side=base_proposal.taker_side,
+        )
+
+        # Get the function result from the base contract
+        neutralizing_contract = base_proposal.neutralize(neutralizing_price, neutralizing_timestamp)
+
+        # Confirm the function result matches expected
+        self.assertEqual(neutralizing_contract, expected_neutralizing_contract)
+
+
 class TestContractProposalValidation(unittest.TestCase):
     def test_ValueError_if_duration_too_short(self):
         with self.assertRaises(ValueError):
@@ -229,7 +286,7 @@ class TestContractProposalValidation(unittest.TestCase):
 FEE_100000_TAKER_TO_MAKER = FeeAgreement(name='fee type a', amount_sats=Sats(100000), receiving=Role.MAKER, paying=Role.TAKER)
 FEE_200000_TAKER_TO_MAKER = FeeAgreement(name='fee type b', amount_sats=Sats(200000), receiving=Role.MAKER, paying=Role.TAKER)
 FEE_400000_TAKER_TO_SETTLEMENT_SERVICE = FeeAgreement(name='fee type ss', amount_sats=Sats(400000), receiving=Role.SETTLEMENT_SERVICE, paying=Role.TAKER)
-FEES = [FEE_100000_TAKER_TO_MAKER, FEE_200000_TAKER_TO_MAKER, FEE_400000_TAKER_TO_SETTLEMENT_SERVICE]
+FEES = (FEE_100000_TAKER_TO_MAKER, FEE_200000_TAKER_TO_MAKER, FEE_400000_TAKER_TO_SETTLEMENT_SERVICE)
 
 
 # Group them for reference
@@ -307,10 +364,10 @@ BEFORE_MATURITY_TIMESTAMP = ScriptTimestamp(MATURITY_TIMESTAMP - 1)
 
 # A valid maturity price, just somewhat over the start price so no chance of low liquidation
 NAIVE_MATURITY_PRICE_USDEM2_PER_BCH = ScriptPriceInOracleUnitsPerBch(220_00)  # $220.00
-BELOW_LIQUIDATING_PRICE_USDEM2_PER_BCH = ScriptPriceInOracleUnitsPerBch(LONG_LIQUIDATION_PRICE - 10)
+BELOW_LIQUIDATING_PRICE_USDEM2_PER_BCH = ScriptPriceInOracleUnitsPerBch(LOW_LIQUIDATION_PRICE - 10)
 
 
-class TestMaturityRedemption(unittest.TestCase):
+class TestRedemption(unittest.TestCase):
     def test_matures_at_maturity_time(self):
         redemption = standard_contract_funding().redeem(
             price_timestamp=MATURITY_TIMESTAMP,
@@ -322,7 +379,7 @@ class TestMaturityRedemption(unittest.TestCase):
     def test_matures_at_maturity_time_even_if_liquidating_price(self):
         redemption = standard_contract_funding().redeem(
             price_timestamp=MATURITY_TIMESTAMP,
-            price_oracleUnits_per_bch=LONG_LIQUIDATION_PRICE,
+            price_oracleUnits_per_bch=LOW_LIQUIDATION_PRICE,
             force_maturity=False,
         )
         self.assertEqual(redemption.redemption_type, RedemptionType.MATURATION)
@@ -338,14 +395,14 @@ class TestMaturityRedemption(unittest.TestCase):
     def test_liquidates_before_maturity_time_at_low_liquidation_price(self):
         redemption = standard_contract_funding().redeem(
             price_timestamp=BEFORE_MATURITY_TIMESTAMP,
-            price_oracleUnits_per_bch=LONG_LIQUIDATION_PRICE,
+            price_oracleUnits_per_bch=LOW_LIQUIDATION_PRICE,
             force_maturity=False,
         )
         self.assertEqual(redemption.redemption_type, RedemptionType.LIQUIDATION)
 
     def test_fails_to_liquidate_before_maturity_time_at_one_over_low_liquidation_price(self):
         valid_liquidation_timestamp = ScriptTimestamp(MATURITY_TIMESTAMP - 1)
-        one_over_liquidation_price = ScriptPriceInOracleUnitsPerBch(LONG_LIQUIDATION_PRICE + 1)
+        one_over_liquidation_price = ScriptPriceInOracleUnitsPerBch(LOW_LIQUIDATION_PRICE + 1)
         with self.assertRaises(UnredeemableError):
             standard_contract_funding().redeem(
                 price_timestamp=valid_liquidation_timestamp,
@@ -362,6 +419,26 @@ class TestMaturityRedemption(unittest.TestCase):
         )
         self.assertEqual(redemption.redemption_type, RedemptionType.MATURATION)
 
+    def test_mutual_redemption_works_if_indicated(self):
+        too_early_maturation_timestamp = ScriptTimestamp(MATURITY_TIMESTAMP - 1)
+        redemption = standard_contract_funding().redeem(
+            price_timestamp=too_early_maturation_timestamp,
+            price_oracleUnits_per_bch=NAIVE_MATURITY_PRICE_USDEM2_PER_BCH,
+            force_maturity=False,
+            is_mutual_redemption=True,
+        )
+        self.assertEqual(redemption.redemption_type, RedemptionType.MUTUAL)
+
+    def test_mutual_redemption_takes_precedence_over_forced_redemption(self):
+        too_early_maturation_timestamp = ScriptTimestamp(MATURITY_TIMESTAMP - 1)
+        redemption = standard_contract_funding().redeem(
+            price_timestamp=too_early_maturation_timestamp,
+            price_oracleUnits_per_bch=NAIVE_MATURITY_PRICE_USDEM2_PER_BCH,
+            force_maturity=True,
+            is_mutual_redemption=True,
+        )
+        self.assertEqual(redemption.redemption_type, RedemptionType.MUTUAL)
+
 
 ###################
 # Redeemed Contract Tests
@@ -369,8 +446,8 @@ class TestMaturityRedemption(unittest.TestCase):
 # Derived short payout sats and units
 # For pure hedge, calculated as max(dust, HEDGE_COMPOSITE // CLAMPED_PRICE - 0) (Note the integer division //)
 SHORT_PAYOUT_SATS_AS_HEDGE = UtxoSats(45_454_545)  # 0.45454545... Bch
-# For leveraged short, same calculation but the cost at short liquidation is not zero
-# Calculated as SHORT_PAYOUT_SATS_AS_HEDGE - COST_SATS_FOR_NOMINAL_VALUE_AT_LEVERAGED_SHORT_LIQUIDATION
+# For leveraged short, same calculation but the cost at high liquidation is not zero
+# Calculated as SHORT_PAYOUT_SATS_AS_HEDGE - COST_SATS_FOR_NOMINAL_VALUE_AT_LEVERAGED_HIGH_LIQUIDATION
 # Due to integer math, this *can* be zero like a perfect hedge in some extreme values.
 SHORT_PAYOUT_SATS_AS_LEVERAGED_SHORT = UtxoSats(18_182_066)
 
@@ -389,27 +466,34 @@ TOTAL_PAYOUT_SATS_AS_HEDGE = UtxoSats(64283878)
 # Calculated as SHORT_PAYOUT_ORACLE_UNITS_AS_HEDGE + LONG_PAYOUT_ORACLE_UNITS
 TOTAL_PAYOUT_ORACLE_UNITS_AS_HEDGE = UsdEM2Beta(14142.4532)
 
+# Nominal value at payout
+# Calculated as round_half_up(NOMINAL_VALUE_USDEM2 / float(CLAMPED_PRICE)) * SATS_PER_BCH
+COST_SATS_FOR_NOMINAL_VALUE_AT_REDEMPTION = Sats(45_454_545)
+
 
 def standard_maturation(
-        funding: ContractFunding | None = None
-):
+        funding: ContractFunding | None = None,
+        fee_agreements: Sequence[FeeAgreement] = FEES,
+) -> ContractRedemption:
     funding = funding if funding is not None else standard_contract_funding()
     return funding.redeem(
         price_timestamp=MATURITY_TIMESTAMP,
         price_oracleUnits_per_bch=NAIVE_MATURITY_PRICE_USDEM2_PER_BCH,
         force_maturity=False,
+        fee_agreements=fee_agreements,
     )
 
 
 def standard_liquidation(
     price_timestamp: ScriptTimestamp = MATURITY_TIMESTAMP,
-    price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch = LONG_LIQUIDATION_PRICE,
-    force_maturity: bool = False,
-):
+    price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch = LOW_LIQUIDATION_PRICE,
+    fee_agreements: Sequence[FeeAgreement] = FEES,
+) -> ContractRedemption:
     return standard_contract_funding().redeem(
         price_timestamp=price_timestamp,
         price_oracleUnits_per_bch=price_oracleUnits_per_bch,
-        force_maturity=force_maturity,
+        force_maturity=False,
+        fee_agreements=fee_agreements,
     )
 
 
@@ -511,6 +595,58 @@ class TestNewRedemptionDerivedValues(unittest.TestCase):
         self.assertAlmostEqual(long_maker_contract.short_gain_oracleUnits, expected_gain_oracleUnits_to_short_taker_at_redemption)
         self.assertAlmostEqual(long_maker_contract.taker_gain_oracleUnits, expected_gain_oracleUnits_to_short_taker_at_redemption)
 
+    ######################
+    # Fee values
+    ######################
+    def test_fee_sats_to_side(self):
+        # Confirm short maker and long taker
+        short_maker_redemption = standard_maturation(standard_contract_funding(standard_contract_proposal(maker_side=Side.SHORT)))
+        long_maker_redemption = standard_maturation(standard_contract_funding(standard_contract_proposal(maker_side=Side.LONG)))
+
+        expected_fees_sats_to_maker = TOTAL_FEES_TO_MAKER_SATS
+        expected_fees_sats_to_taker = TOTAL_FEES_TO_TAKER_SATS
+
+        # Short Maker
+        self.assertEqual(short_maker_redemption.fee_sats_to_short, expected_fees_sats_to_maker)
+        self.assertEqual(short_maker_redemption.fee_sats_to_maker, expected_fees_sats_to_maker)
+
+        # Long Taker
+        self.assertEqual(short_maker_redemption.fee_sats_to_long, expected_fees_sats_to_taker)
+        self.assertEqual(short_maker_redemption.fee_sats_to_taker, expected_fees_sats_to_taker)
+
+        # Long Maker
+        self.assertEqual(long_maker_redemption.fee_sats_to_long, expected_fees_sats_to_maker)
+        self.assertEqual(long_maker_redemption.fee_sats_to_maker, expected_fees_sats_to_maker)
+
+        # Short Taker
+        self.assertEqual(long_maker_redemption.fee_sats_to_short, expected_fees_sats_to_taker)
+        self.assertEqual(long_maker_redemption.fee_sats_to_taker, expected_fees_sats_to_taker)
+
+    def test_fee_oracleUnits_to_side(self):
+        # Confirm short maker and long taker
+        short_maker_redemption = standard_maturation(standard_contract_funding(standard_contract_proposal(maker_side=Side.SHORT)))
+        long_maker_redemption = standard_maturation(standard_contract_funding(standard_contract_proposal(maker_side=Side.LONG)))
+        oracle_class = short_maker_redemption.base_funding.base_proposal.oracle_unit_cls
+
+        expected_fee_oracleUnits_to_maker_at_redemption = oracle_class(TOTAL_FEES_TO_MAKER_SATS * NAIVE_MATURITY_PRICE_USDEM2_PER_BCH / SATS_PER_BCH)
+        expected_fee_oracleUnits_to_taker_at_redemption = oracle_class(TOTAL_FEES_TO_TAKER_SATS * NAIVE_MATURITY_PRICE_USDEM2_PER_BCH / SATS_PER_BCH)
+
+        # Short Maker
+        self.assertEqual(short_maker_redemption.fee_oracleUnits_to_short, expected_fee_oracleUnits_to_maker_at_redemption)
+        self.assertEqual(short_maker_redemption.fee_oracleUnits_to_maker, expected_fee_oracleUnits_to_maker_at_redemption)
+
+        # Long Taker
+        self.assertEqual(short_maker_redemption.fee_oracleUnits_to_long, expected_fee_oracleUnits_to_taker_at_redemption)
+        self.assertEqual(short_maker_redemption.fee_oracleUnits_to_taker, expected_fee_oracleUnits_to_taker_at_redemption)
+
+        # Long Maker
+        self.assertEqual(long_maker_redemption.fee_oracleUnits_to_long, expected_fee_oracleUnits_to_maker_at_redemption)
+        self.assertEqual(long_maker_redemption.fee_oracleUnits_to_maker, expected_fee_oracleUnits_to_maker_at_redemption)
+
+        # Short Taker
+        self.assertEqual(long_maker_redemption.fee_oracleUnits_to_short, expected_fee_oracleUnits_to_taker_at_redemption)
+        self.assertEqual(long_maker_redemption.fee_oracleUnits_to_taker, expected_fee_oracleUnits_to_taker_at_redemption)
+
     ######################
     # Sided versions of values
     ######################
@@ -574,12 +710,12 @@ class TestHedgeVsLeveragedShort(unittest.TestCase):
         # Confirm that the new cost at start is correct
         self.assertEqual(self.leveraged_short.cost_sats_for_nominal_value_at_start, COST_SATS_FOR_NOMINAL_VALUE_AT_START)
 
-    def test_short_liquidation_price_for_leveraged_short(self):
-        self.assertEqual(self.leveraged_short.short_liquidation_price_oracleUnits_per_bch, SHORT_LIQUIDATION_PRICE_AS_LEVERAGED_SHORT)
+    def test_high_liquidation_price_for_leveraged_short(self):
+        self.assertEqual(self.leveraged_short.high_liquidation_price_oracleUnits_per_bch, HIGH_LIQUIDATION_PRICE_AS_LEVERAGED_SHORT)
 
-    def test_cost_at_short_liquidation_for_hedge_vs_leveraged_short(self):
-        self.assertEqual(self.hedge_short.cost_sats_for_nominal_value_at_short_liquidation, ZERO_COST_OF_NOMINAL_VALUE_AT_HEDGE_LIQUIDATION)
-        self.assertEqual(self.leveraged_short.cost_sats_for_nominal_value_at_short_liquidation, COST_SATS_FOR_NOMINAL_VALUE_AT_LEVERAGED_SHORT_LIQUIDATION)
+    def test_cost_at_high_liquidation_for_hedge_vs_leveraged_short(self):
+        self.assertEqual(self.hedge_short.cost_sats_for_nominal_value_at_high_liquidation, ZERO_COST_OF_NOMINAL_VALUE_AT_HEDGE_LIQUIDATION)
+        self.assertEqual(self.leveraged_short.cost_sats_for_nominal_value_at_high_liquidation, COST_SATS_FOR_NOMINAL_VALUE_AT_LEVERAGED_HIGH_LIQUIDATION)
 
     def test_short_input_sats_for_leveraged_short(self):
         self.assertEqual(self.leveraged_short.short_input_sats, SHORT_INPUT_SATS_AS_LEVERAGED_SHORT)
@@ -598,6 +734,24 @@ class TestHedgeVsLeveragedShort(unittest.TestCase):
         leveraged_short_redemption = standard_maturation(standard_contract_funding(self.leveraged_short))
         self.assertEqual(leveraged_short_redemption.short_payout_sats, SHORT_PAYOUT_SATS_AS_LEVERAGED_SHORT)
 
+    def test_cost_at_redemption_for_hedge_vs_leveraged_short(self):
+        hedge_redemption = standard_maturation(standard_contract_funding(self.hedge_short))
+        leveraged_short_redemption = standard_maturation(standard_contract_funding(self.leveraged_short))
+
+        # hedge redemption should just be the short payout
+        self.assertEqual(hedge_redemption.cost_sats_for_nominal_value_at_redemption, hedge_redemption.short_payout_sats)
+
+        # leveraged short redemption is different from short payout
+        self.assertNotEqual(leveraged_short_redemption.cost_sats_for_nominal_value_at_redemption, leveraged_short_redemption.short_payout_sats)
+        self.assertEqual(leveraged_short_redemption.cost_sats_for_nominal_value_at_redemption, COST_SATS_FOR_NOMINAL_VALUE_AT_REDEMPTION)
+
+
+class TestContractSide(unittest.TestCase):
+    def test_from_string_works_for_known_api_values(self):
+        self.assertEqual(Side.from_string('short'), Side.SHORT)
+        self.assertEqual(Side.from_string('hedge'), Side.SHORT)
+        self.assertEqual(Side.from_string('long'), Side.LONG)
+
 
 if __name__ == '__main__':
     unittest.main()
